"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const sequelize_1 = require("sequelize");
const versioning_1 = require("../utils/versioning");
const string_1 = require("../utils/string");
const models_1 = require("../services/models");
const object_1 = require("../utils/object");
const parentPrototype = versioning_1.majorVersion === 3 ? sequelize_1.Instance.prototype : sequelize_1.Model.prototype;
class BaseModel {
    static extend(target) {
        object_1.extend(target, this);
        overrideStaticFunctions(target);
        /**
         * Overrides all static functions with a function, that
         * checks if corresponding model is initialized and
         * prepares given options if necessary
         */
        function overrideStaticFunctions(_target) {
            const isFunctionMember = key => typeof _target[key] !== 'function';
            object_1.getAllPropertyNames(_target)
                .filter(key => !canOverrideMember(key))
                .forEach(key => {
                if (isFunctionMember(key))
                    return;
                const superFn = _target[key];
                _target[key] = function (...args) {
                    checkInitialization(this, key);
                    tryPrepareOptions(this, key, args);
                    return superFn.call(this, ...args);
                };
            });
        }
        /**
         * Checks if member - specified by propertyKey - can be overridden or not
         */
        function canOverrideMember(propertyKey) {
            if (isPrivateMember(propertyKey)) {
                return true;
            }
            const FORBIDDEN_KEYS = ['name', 'constructor', 'length', 'prototype', 'caller', 'arguments', 'apply',
                'QueryInterface', 'QueryGenerator', 'init', 'replaceHookAliases', 'refreshAttributes'];
            return FORBIDDEN_KEYS.indexOf(propertyKey) !== -1;
        }
        /**
         * Checks if member is private or not. Is identified by starting
         * "_" in specified key
         */
        function isPrivateMember(propertyKey) {
            return (propertyKey.charAt(0) === '_');
        }
        /**
         * Checks if model is initialized
         * @throw if model is not initialized
         */
        function checkInitialization(model, propertyKey) {
            if (!model.isInitialized) {
                throw new Error(`Model not initialized: "${model.name}" needs to be added to a Sequelize instance ` +
                    `before "${propertyKey}" can be called.`);
            }
        }
        /**
         * Prepares options if necessary:
         *  - infers alias of given options
         */
        function tryPrepareOptions(model, propertyKey, args) {
            const optionIndex = models_1.INFER_ALIAS_MAP[propertyKey];
            if (optionIndex !== undefined) {
                const options = args[optionIndex];
                if (options) {
                    args[optionIndex] = models_1.inferAlias(options, model);
                }
            }
        }
    }
    /**
     * Prepares build options for instantiation of a model
     */
    static prepareInstantiationOptions(options, source) {
        options = models_1.inferAlias(options, source);
        if (!('isNewRecord' in options))
            options.isNewRecord = true;
        if (!('$schema' in options) && this['$schema'])
            options['$schema'] = this['$schema'];
        if (!('$schemaDelimiter' in options) && this['$schemaDelimiter'])
            options['$schemaDelimiter'] = this['$schemaDelimiter'];
        const staticMethodPrefix = versioning_1.majorVersion === 3 ? '$' : '_';
        // preventing TypeError: Cannot read property 'indexOf' of undefined(=includeNames)
        if (!options['includeNames'])
            options['includeNames'] = [];
        if (!options['includeValidated']) {
            sequelize_1.Model[staticMethodPrefix + 'conformOptions'](options, source);
            if (options.include) {
                sequelize_1.Model[staticMethodPrefix + 'expandIncludeAll'].call(source, options);
                sequelize_1.Model[staticMethodPrefix + 'validateIncludedElements'].call(source, options);
            }
        }
        return options;
    }
    /**
     * Adds relation between specified instances and source instance
     */
    $add(propertyKey, instances, options) {
        return this['add' + string_1.capitalize(propertyKey)](instances, options);
    }
    ;
    /**
     * Sets relation between specified instances and source instance
     * (replaces old relations)
     */
    $set(propertyKey, instances, options) {
        return this['set' + string_1.capitalize(propertyKey)](instances, options);
    }
    ;
    /**
     * Returns related instance (specified by propertyKey) of source instance
     */
    $get(propertyKey, options) {
        return this['get' + string_1.capitalize(propertyKey)](options);
    }
    ;
    /**
     * Counts related instances (specified by propertyKey) of source instance
     */
    $count(propertyKey, options) {
        return this['count' + string_1.capitalize(propertyKey)](options);
    }
    ;
    /**
     * Creates instances and relate them to source instance
     */
    $create(propertyKey, values, options) {
        return this['create' + string_1.capitalize(propertyKey)](values, options);
    }
    ;
    /**
     * Checks if specified instances is related to source instance
     */
    $has(propertyKey, instances, options) {
        return this['has' + string_1.capitalize(propertyKey)](instances, options);
    }
    ;
    /**
     * Removes specified instances from source instance
     */
    $remove(propertyKey, instances, options) {
        return this['remove' + string_1.capitalize(propertyKey)](instances, options);
    }
    ;
    /**
     * Overridden due to infer alias from options is required
     *
     * SEE DETAILS FOR ACTUAL FUNCTIONALITY ON DECLARATION FILE
     */
    reload(options) {
        return parentPrototype.reload.call(this, models_1.inferAlias(options, this));
    }
    ;
}
BaseModel.isInitialized = false;
exports.BaseModel = BaseModel;
//# sourceMappingURL=BaseModel.js.map