"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
require("reflect-metadata");
const lodash_1 = require("lodash");
exports.BELONGS_TO_MANY = 'belongsToMany';
exports.BELONGS_TO = 'belongsTo';
exports.HAS_MANY = 'hasMany';
exports.HAS_ONE = 'hasOne';
const FOREIGN_KEYS_KEY = 'sequelize:foreignKeys';
const ASSOCIATIONS_KEY = 'sequelize:associations';
/**
 * Stores association meta data for specified class
 */
function addAssociation(target, relation, relatedClassGetter, as, optionsOrForeignKey, through, otherKey) {
    let associations = getAssociations(target);
    if (!associations) {
        associations = [];
        setAssociations(target, associations);
    }
    let throughClassGetter;
    if (typeof through === 'function') {
        throughClassGetter = through;
        through = undefined;
    }
    let options = {};
    if (typeof optionsOrForeignKey === 'string') {
        options.foreignKey = { name: optionsOrForeignKey };
    }
    else {
        options = Object.assign({}, optionsOrForeignKey);
    }
    if (otherKey) {
        options.otherKey = { name: otherKey };
    }
    associations.push({
        relation,
        relatedClassGetter,
        throughClassGetter,
        through: through,
        as,
        options
    });
}
exports.addAssociation = addAssociation;
/**
 * Determines foreign key by specified association (relation)
 */
function getForeignKey(model, association) {
    const options = association.options;
    if (options && options.foreignKey) {
        const foreignKey = options.foreignKey;
        // if options is an object and has a string foreignKey property, use that as the name
        if (typeof foreignKey === 'string') {
            return foreignKey;
        }
        // if options is an object with foreignKey.name, use that as the name
        if (foreignKey.name) {
            return foreignKey.name;
        }
    }
    // otherwise calculate the foreign key by related or through class
    let classWithForeignKey;
    let relatedClass;
    switch (association.relation) {
        case exports.BELONGS_TO_MANY:
            if (association.throughClassGetter) {
                classWithForeignKey = association.throughClassGetter();
                relatedClass = model;
            }
            else {
                throw new Error(`ThroughClassGetter is missing on "${model['name']}"`);
            }
            break;
        case exports.HAS_MANY:
        case exports.HAS_ONE:
            classWithForeignKey = association.relatedClassGetter();
            relatedClass = model;
            break;
        case exports.BELONGS_TO:
            classWithForeignKey = model;
            relatedClass = association.relatedClassGetter();
            break;
        default:
    }
    const foreignKeys = getForeignKeys(classWithForeignKey.prototype) || [];
    for (const foreignKey of foreignKeys) {
        if (foreignKey.relatedClassGetter() === relatedClass) {
            return foreignKey.foreignKey;
        }
    }
    throw new Error(`Foreign key for "${relatedClass.name}" is missing ` +
        `on "${classWithForeignKey.name}".`);
}
exports.getForeignKey = getForeignKey;
/**
 * Returns association meta data from specified class
 */
function getAssociations(target) {
    return Reflect.getMetadata(ASSOCIATIONS_KEY, target);
}
exports.getAssociations = getAssociations;
function setAssociations(target, associations) {
    Reflect.defineMetadata(ASSOCIATIONS_KEY, associations, target);
}
exports.setAssociations = setAssociations;
function getAssociationsByRelation(target, relatedClass) {
    const associations = getAssociations(target);
    return (associations || []).filter(association => {
        const _relatedClass = association.relatedClassGetter();
        return (_relatedClass.prototype === relatedClass.prototype ||
            /* istanbul ignore next */
            relatedClass.prototype instanceof _relatedClass // v4 (for child classes)
        );
    });
}
exports.getAssociationsByRelation = getAssociationsByRelation;
/**
 * Adds foreign key meta data for specified class
 */
function addForeignKey(target, relatedClassGetter, foreignKey) {
    let foreignKeys = getForeignKeys(target);
    if (!foreignKeys) {
        foreignKeys = [];
        setForeignKeys(target, foreignKeys);
    }
    foreignKeys.push({
        relatedClassGetter,
        foreignKey
    });
}
exports.addForeignKey = addForeignKey;
/**
 * Returns "other" key determined by association object
 */
function getOtherKey(association) {
    const options = association.options;
    if (options && options.otherKey) {
        const otherKey = options.otherKey;
        // if options is an object and has a string otherKey property, use that as the name
        if (typeof otherKey === 'string') {
            return otherKey;
        }
        // if options is an object with otherKey.name, use that as the name
        if (otherKey.name) {
            return otherKey.name;
        }
    }
    return getForeignKey(association.relatedClassGetter(), association);
}
exports.getOtherKey = getOtherKey;
/**
 * Processes association for single model
 */
function processAssociation(sequelize, model, association) {
    const relatedClass = association.relatedClassGetter();
    const foreignKey = getForeignKey(model, association);
    let through;
    let otherKey;
    if (association.relation === exports.BELONGS_TO_MANY) {
        otherKey = getOtherKey(association);
        through = getThroughClass(sequelize, association);
    }
    const foreignKeyOptions = { foreignKey: { name: foreignKey } };
    if (otherKey) {
        foreignKeyOptions.otherKey = { name: otherKey };
    }
    const options = lodash_1.merge(association.options, foreignKeyOptions, {
        as: association.as,
        through,
    });
    model[association.relation](relatedClass, options);
    sequelize.adjustAssociation(model, association);
}
exports.processAssociation = processAssociation;
/**
 * Returns "through" class determined by association object
 */
function getThroughClass(sequelize, association) {
    if (association.through) {
        if (!sequelize.thoughMap[association.through]) {
            const throughModel = sequelize.getThroughModel(association.through);
            sequelize.addModels([throughModel]);
            sequelize.thoughMap[association.through] = throughModel;
        }
        return sequelize.thoughMap[association.through];
    }
    return association.throughClassGetter();
}
exports.getThroughClass = getThroughClass;
/**
 * Returns foreign key meta data from specified class
 */
function getForeignKeys(target) {
    return Reflect.getMetadata(FOREIGN_KEYS_KEY, target);
}
/**
 * Sets foreign key meta data
 */
function setForeignKeys(target, foreignKeys) {
    Reflect.defineMetadata(FOREIGN_KEYS_KEY, foreignKeys, target);
}
//# sourceMappingURL=association.js.map